local addonName, TS = ...; TS.Utils = { }

--[[
	Join 2 arrays into a new single array.
]]
local function Join(arr1, arr2)
	local newArray = { }

	for _, value in pairs(arr1) 
	do
		table.insert(newArray, value)
	end
	
	if (arr2 == nil) then return newArray end
	
	for _, value in pairs(arr2) 
	do
		table.insert(newArray, value)
	end
	
	return newArray
end

--[[
	Join alll the provided arrays into 1. Any parameter past the first can be nil.
]]
local function JoinMany(arr1, arr2, arr3, arr4, arr5, arr6, arr7)
	return Join(Join(Join(Join(Join(Join(arr1, arr2), arr3), arr4), arr5), arr6), arr7)
end

--Each array contains all the spell IDs of all ranks for that totem from low to high
local FIRE_NOVA = { 1535, 8498, 8499, 11314, 11315 }
local STONECLAW = { 5730, 6390, 6391, 6392, 10427, 10428 }
--Each array contains all the aura spell IDs of all ranks for that totem from low to high
local FIRE_RES = { 8185, 10534, 10535 }
local FROST_RES = { 8182, 10476, 10477 }
local GRACE_OF_AIR = { 8836, 10626, 25360 }
local GROUNDING = { 8178 }
local NATURE_RES = { 10596, 10598, 10599 }
local STONESKIN = { 8072, 8156, 8157, 10403, 10404, 10405 }
local STRENGTH_OF_EARTH = { 8076, 8162, 8163, 10441, 25362 }
local WINDWALL = { 15108, 15109, 15110 }
local HEALING_STREAM = { 5672, 6371, 6372, 10460, 10461 }
local MANA_SPRING = { 5677, 10491, 10493, 10494, 24853 }
local MANA_TIDE = { 16191, 17355, 17360 }
local TRANQUIL_AIR = { 25909 }
local ALL_BENEFIT = JoinMany(FIRE_RES, FROST_RES, GROUNDING, NATURE_RES, WINDWALL, HEALING_STREAM, TRANQUIL_AIR)

--Defines which totems are considered useful to which classes
local ClassUsefulTotemAuras = 
{ 
	[1] = JoinMany(ALL_BENEFIT, GRACE_OF_AIR, STONESKIN, STRENGTH_OF_EARTH), --Warrior
	[3] = JoinMany(ALL_BENEFIT, GRACE_OF_AIR, MANA_SPRING, MANA_TIDE), --Hunter
	[4] = JoinMany(ALL_BENEFIT, GRACE_OF_AIR, STONESKIN, STRENGTH_OF_EARTH), --Rogue
	[5] = JoinMany(ALL_BENEFIT, MANA_SPRING, MANA_TIDE), --Priest
	[7] = JoinMany(ALL_BENEFIT, GRACE_OF_AIR, STONESKIN, STRENGTH_OF_EARTH, MANA_SPRING, MANA_TIDE), --Shaman
	[8] = JoinMany(ALL_BENEFIT, MANA_SPRING, MANA_TIDE), --Mage
	[9] = JoinMany(ALL_BENEFIT, MANA_SPRING, MANA_TIDE), --Warlock 
	[11] = JoinMany(ALL_BENEFIT, GRACE_OF_AIR, STONESKIN, STRENGTH_OF_EARTH, MANA_SPRING, MANA_TIDE) --Druid 
}

local function Totem(totemIndex, auraId, health)
	return { TotemIndex = totemIndex, AuraId = auraId, Health = health }
end

TS.Totems = 
{
	[2484] = Totem(1, 0, 5),		--Earthbind
	
	[1535] = Totem(2, 0, 5),		--Fire Nova Rank 1
	[8498] = Totem(2, 0, 5),		--Fire Nova Rank 2
	[8499] = Totem(2, 0, 5),		--Fire Nova Rank 3
	[11314] = Totem(2, 0, 5),		--Fire Nova Rank 4
	[11315] = Totem(2, 0, 5),		--Fire Nova Rank 5
	
	[8190] = Totem(2, 0, 5),		--Magma Rank 1
	[10585] = Totem(2, 0, 5),		--Magma Rank 2
	[10586] = Totem(2, 0, 5),		--Magma Rank 3
	[10587] = Totem(2, 0, 5),		--Magma Rank 4

	[3599] = Totem(2, 0, 5),		--Searing Rank 1
	[6363] = Totem(2, 0, 5),		--Searing Rank 2
	[6364] = Totem(2, 0, 5),		--Searing Rank 3
	[6365] = Totem(2, 0, 5),		--Searing Rank 4
	[10437] = Totem(2, 0, 5),		--Searing Rank 5
	[10438] = Totem(2, 0, 5),		--Searing Rank 6
	
	[5730] = Totem(1, 0, 50),		--Stoneclaw Rank 1
	[6390] = Totem(1, 0, 150),		--Stoneclaw Rank 2
	[6391] = Totem(1, 0, 220),		--Stoneclaw Rank 3
	[6392] = Totem(1, 0, 280),		--Stoneclaw Rank 4
	[10427] = Totem(1, 0, 390),		--Stoneclaw Rank 5
	[10428] = Totem(1, 0, 480),		--Stoneclaw Rank 6
	
	[8184] = Totem(3, 8185, 5),		--Fire Resistance Rank 1
	[10537] = Totem(3, 10534, 5), 	--Fire Resistance Rank 2
	[10538] = Totem(3, 10535, 5), 	--Fire Resistance Rank 3
	
	[8227] = Totem(2, 0, 5),		--Flametongue Rank 1
	[8249] = Totem(2, 0, 5),		--Flametongue Rank 2
	[10526] = Totem(2, 0, 5),  		--Flametongue Rank 3
	[16387] = Totem(2, 0, 5),  		--Flametongue Rank 4
	
	[8181] = Totem(2, 8182, 5),		--Frost Resistance Rank 1
	[10478] = Totem(2, 10476, 5),  	--Frost Resistance Rank 2
	[10479] = Totem(2, 10477, 5),  	--Frost Resistance Rank 3
	
	[8835] = Totem(4, 8836, 5),  	--Grace of Air Rank 1
	[10627] = Totem(4, 10626, 5),  	--Grace of Air Rank 2
	[25359] = Totem(4, 25360, 5),  	--Grace of Air Rank 3
	
	[8177] = Totem(4, 8178, 5), 	--Grounding
	
	[10595] = Totem(4, 10596, 5),  	--Nature Resistance Rank 1
	[10600] = Totem(4, 10598, 5),  	--Nature Resistance Rank 2
	[10601] = Totem(4, 10599, 5),  	--Nature Resistance Rank 3
	
	[6495] = Totem(4, 0, 100),  	--Sentry
	
	[8071] = Totem(1, 8072, 5),  	--Stoneskin Rank 1
	[8154] = Totem(1, 8156, 5),  	--Stoneskin Rank 2
	[8155] = Totem(1, 8157, 5),  	--Stoneskin Rank 3
	[10406] = Totem(1, 10403, 5),  	--Stoneskin Rank 4
	[10407] = Totem(1, 10404, 5),  	--Stoneskin Rank 5
	[10408] = Totem(1, 10405, 5),  	--Stoneskin Rank 6
	
	[8075] = Totem(1, 8076, 5),  	--Strenth of Earth Rank 1
	[8160] = Totem(1, 8162, 5),  	--Strenth of Earth Rank 2
	[8161] = Totem(1, 8163, 5),  	--Strenth of Earth Rank 3
	[10442] = Totem(1, 10441, 5),  	--Strenth of Earth Rank 4
	[25361] = Totem(1, 25362, 5),  	--Strenth of Earth Rank 5

	[8512] = Totem(4, 0, 5) ,  		--Windfury Rank 1
	[10613] = Totem(4, 0, 5) ,  	--Windfury Rank 2
	[10614] = Totem(4, 0, 5) ,  	--Windfury Rank 3
	
	[15107] = Totem(4, 15108, 5),  	--Windwall Rank 1
	[15111] = Totem(4, 15109, 5),  	--Windwall Rank 2
	[15112] = Totem(4, 15110, 5),  	--Windwall Rank 3
	
	[8170] = Totem(3, 0, 5),  		--Disease Cleansing
	
	[5394] = Totem(3, 5672, 5),  	--Healing Stream Rank 1
	[6375] = Totem(3, 6371, 5),  	--Healing Stream Rank 2
	[6377] = Totem(3, 6372, 5),  	--Healing Stream Rank 3
	[10462] = Totem(3, 10460, 5),  	--Healing Stream Rank 4
	[10463] = Totem(3, 10461, 5),  	--Healing Stream Rank 5
	
	[5675] = Totem(3, 5677, 5),  	--Mana Spring Rank 1
	[10495] = Totem(3, 10491, 5),  	--Mana Spring Rank 2
	[10496] = Totem(3, 10493, 5),  	--Mana Spring Rank 3
	[10497] = Totem(3, 10494, 5),  	--Mana Spring Rank 4

	[16190] = Totem(3, 16191, 5),  	--Mana Tide Rank 1
	[17354] = Totem(3, 17355, 5),  	--Mana Tide Rank 2
	[17359] = Totem(3, 17360, 5),  	--Mana Tide Rank 3
	
	[8166] = Totem(3, 0, 5),  		--Poison Cleansing

	[25908] = Totem(4, 25909, 5),  	--Tranquil Air
	
	[8143] = Totem(1, 0, 5), 		--Tremor
	
	[24854] = Totem(3, 24853, 5)	--Enamored Water Spirit (Trinket)
}

--[[
	Updates health data of stoneclaw totem (determined by variables such as player level and talents).
	Should be called when either of these 2 change.
]]
function TS.Utils.UpdateVariableData()
	--get player level, rank of Earth's Grasp talent and set the bonus multiplier based on rank
	local playerLevel = UnitLevel("player")
	local a, b, c, d, earthsGraspRank = GetTalentInfo(1, 3)
	local earthsGraspBonus = 1 + (0.25 * earthsGraspRank)	
	
	--loop over all stoneclaw ranks 1-6 starting from 1
	local levelGainedAt = 8
	for _, spellId in pairs(STONECLAW) 
	do
		local baseHealth = TS.Totems[spellId].Health
		local levelBonus = (playerLevel - levelGainedAt) * 3
		TS.Totems[spellId].Health = (baseHealth + levelBonus) * earthsGraspBonus

		levelGainedAt = levelGainedAt + 10
	end
end

--[[
	Determines if the totem aura with the given aura spell ID 
	is useful to the unit with the given unit ID.
]]
function TS.Utils.IsAuraUsefulToUnit(unitId, spellId)
	local a, b, classId = UnitClass(unitId)
	
	--find the useful auras for this class
	local usefulAuras = ClassUsefulTotemAuras[classId]
	if (usefulAuras == nil) then return end
	
	--loop over all the values to see if it contains our spell ID
	for _, value in pairs(usefulAuras) do
		if value == spellId then return true end
	end
	
	return false
end

function TS.Utils.GetSpellTotemIndex(spellId)
	if (TS.Totems[spellId] == nil) then return nil end
	
	return TS.Totems[spellId].TotemIndex
end

--[[
	Get the unit ID string used by the Blizzard API for the given unit index (1-5).
	1 = player, 2, 3, 4, 5 = party members
]]
function TS.Utils.GetUnitId(unitIndex)
	if (unitIndex == 1) then return "player" end
	
	return "party" .. (unitIndex - 1)
end

--[[
	Determine if the unit with the given ID is affected with the given spell ID.
]]
function TS.Utils.IsUnitAffected(unitId, spellId)
	for buffIndex = 1, 40 
	do
		--get the spell ID for this buff index
		local a, b, c, d, e, f, g, h, i, unitSpellId = UnitAura(unitId, buffIndex, "HELPFUL")
		if (unitSpellId == spellId) then return true end
	end
	
	return false
end

--[[
	Determine if the unit is able to receive totem buffs.
]]
function TS.Utils.CanUnitReceiveTotemBuffs(unitId)
    return (UnitIsConnected(unitId) and UnitExists(unitId) and not UnitIsDeadOrGhost(unitId))
end

--[[
	Determines if the totem with the provided spell ID should be ignored in the expiry alarm.
]]
function TS.Utils.IsTotemIgnoredForAlarm(spellId)
	for index, value in ipairs(FIRE_NOVA) do
        if spellId == value then return true end
    end

    return false
end

local function IsTotemNameInArray(array, totemNameToFind)
    for _, totemName in pairs(array) do
		if (string.find(totemNameToFind, totemName)) then return true end
    end
	
    return false
end

local _earthTotems = { "Earthbind", "Stoneclaw", "Stoneskin", "Strenth of Earth", "Tremor" }
local _fireTotems = { "Fire Nova", "Magma", "Searing", "Flametongue", "Frost Resistance" }
local _waterTotems = { "Fire Resistance", "Disease Cleansing", "Healing Stream", "Mana Spring", "Mana Tide", "Poison Cleansing", "Ancient Mana Spring" }
local _airTotems = { "Grace of Air", "Grounding", "Nature Resistance", "Sentry", "Windfury", "Windwall", "Tranquil Air" }

function TS.Utils.GetTotemIndexFromName(totemName)
	if (totemName == nil or string.find(totemName, "Totem") == false) then return nil end
	
	if (IsTotemNameInArray(_earthTotems, totemName)) then return 1 end
	if (IsTotemNameInArray(_fireTotems, totemName)) then return 2 end
	if (IsTotemNameInArray(_waterTotems, totemName)) then return 3 end
	if (IsTotemNameInArray(_airTotems, totemName)) then return 4 end

	return nil
end